#ifndef GPUX
#define GPUX

// -------------------------------------------------------------------------------------------------
// Pragmas
// -------------------------------------------------------------------------------------------------

// Use row-major layout for storing matrix values.
//
// By convention, shader code uses the product "matrix * column-vector", which will make the
// compiler output DP4 instructions (for row-major layout), instead of MUL/MAD (for column-major
// layout).
//
// The product "row-vector * matrix" is only used as an inplace optimization for
// "transpose(matrix) * column-vector".
//
// Changing the layout to column-major will not break the behaviour of the shader code. However, the
// code that uploads matrix values to the GPU must be aware of the memory layout, which is the case
// for the render effect framework of Tinman 3D.
//
// With row-major layout, the 'float3x4' type consumes three consecutive 'float4' values.
#pragma pack_matrix(row_major)

#ifdef GX_HLSL_DX9

// X3206: implicit truncation of vector type
#pragma warning (disable : 3206)

#endif

#ifndef GX_HLSL_DX12

// X3550: sampler array index must be a literal expression, forcing loop to unroll.
#pragma warning (disable : 3550)

// X3556: integer divides may be much slower, try using uints if possible.
#pragma warning (disable : 3556)

// X3571: pow(f, e) will not work for negative f, use abs(f) or conditionally handle negative values
#pragma warning (disable : 3571)

#endif

// -------------------------------------------------------------------------------------------------
// Macros
// -------------------------------------------------------------------------------------------------

#define SEMANTIC_COLOR(INDEX) : COLOR##INDEX
#define SEMANTIC_POSITION(INDEX) : POSITION##INDEX
#define SEMANTIC_NORMAL(INDEX) : NORMAL##INDEX
#define SEMANTIC_TEXCOORD(INDEX) : TEXCOORD##INDEX
#define SEMANTIC_TANGENT(INDEX) : TANGENT##INDEX

#ifdef GX_HLSL_DX9 // -----------------------------------------------------------------------------

#define SEMANTIC_SYSTEM_POSITION(INDEX) : POSITION##INDEX
#define SEMANTIC_SYSTEM_TARGET(INDEX) : COLOR##INDEX
#define SEMANTIC_SYSTEM_DEPTH : DEPTH
#define SEMANTIC_SYSTEM_DEPTH_LESS_EQUAL : DEPTH
#define SEMANTIC_SYSTEM_DEPTH_GREATER_EQUAL : DEPTH
#define SEMANTIC_SYSTEM_DISPATCH_ID __NOT_SUPPORTED__
#define SEMANTIC_SYSTEM_THREAD_ID __NOT_SUPPORTED__
#define SEMANTIC_SYSTEM_TESSELLATE_FACTOR_INNER __NOT_SUPPORTED__
#define SEMANTIC_SYSTEM_TESSELLATE_FACTOR_OUTER __NOT_SUPPORTED__
#define SEMANTIC_CUSTOM(NAME) // Unsupported, skip declaration

#define ARRAY(NAME) // Unsupported, skip declaration
#define ARRAY_RW(NAME) // Unsupported, skip declaration
#define BUFFER(TYPE,NAME) // Unsupported, skip declaration
#define BUFFER_RW(TYPE,NAME) // Unsupported, skip declaration
#define TEXTURE_2D(TYPE,NAME) sampler2D NAME
#define TEXTURE_2D_ARRAY(TYPE,NAME) sampler2D NAME
#define TEXTURE_2D_RW(TYPE,NAME) // Unsupported, skip declaration
#define TEXTURE_CUBE(TYPE,NAME) samplerCUBE NAME

#define gxSizeOf(T,S) __NOT_SUPPORTED__
#define gxSizeOfLevel(T,L,S) 0 // Allow compilation in disabled conditional branches.
#define gxStreamAppend(S,V) __NOT_SUPPORTED__
#define gxStreamReset(S) __NOT_SUPPORTED__
#define gxSync __NOT_SUPPORTED__
#define gxTexture2D(T,C) tex2D(T,C)
#define gxTexture2DArray(T,S,C) tex2D(T,C)
#define gxTexture2DArrayLod(T,S,C,L) tex2Dlod(T,float4(C,0,L))
#define gxTexture2DGet(V,I) __NOT_SUPPORTED__
#define gxTexture2DLod(T,C,L) tex2Dlod(T,float4(C,0,L))
#define gxTexture2DSet(V,I,X) __NOT_SUPPORTED__
#define gxTextureCube(T,C) texCUBE(T,C)
#define gxTextureCubeLod(T,C,L) texCUBElod(T,float4(C,L))
#define gxTextures2D(T,N,C) tex2D(T[N],C)
#define gxTextures2DArray(T,N,S,C) tex2D(T[N],C)
#define gxTextures2DArrayLod(T,N,S,C,L) tex2Dlod(T[N],float4(C,0,L))
#define gxTextures2DLod(T,N,C,L) tex2Dlod(T[N],float4(C,0,L))
#define gxTexturesCube(T,N,C) texCUBE(T[N],C)
#define gxTexturesCubeLod(T,N,C,L) texCUBElod(T[N],float4(C,L))

#else // ------------------------------------------------------------------------------------------

#define SEMANTIC_SYSTEM_POSITION(INDEX) : SV_Position##INDEX
#define SEMANTIC_SYSTEM_TARGET(INDEX) : SV_Target##INDEX
#define SEMANTIC_SYSTEM_DEPTH : SV_Depth
#define SEMANTIC_SYSTEM_DEPTH_LESS_EQUAL : SV_DepthLessEqual
#define SEMANTIC_SYSTEM_DEPTH_GREATER_EQUAL : SV_DepthGreaterEqual
#define SEMANTIC_SYSTEM_DISPATCH_ID : SV_GroupID
#define SEMANTIC_SYSTEM_THREAD_ID : SV_GroupThreadID
#define SEMANTIC_SYSTEM_TESSELLATE_FACTOR_INNER : SV_InsideTessFactor
#define SEMANTIC_SYSTEM_TESSELLATE_FACTOR_OUTER : SV_TessFactor
#define SEMANTIC_CUSTOM(NAME) : NAME

#define ARRAY(NAME) Buffer<uint> NAME
#define ARRAY_RW(NAME) RWBuffer<uint> NAME
#define BUFFER(TYPE,NAME) StructuredBuffer<TYPE> NAME
#define BUFFER_RW(TYPE,NAME) RWStructuredBuffer<TYPE> NAME
#define TEXTURE_2D(TYPE,NAME) SamplerState NAME##Sampler;Texture2D<TYPE> NAME
#define TEXTURE_2D_ARRAY(TYPE,NAME) SamplerState NAME##Sampler;Texture2DArray<TYPE> NAME
#define TEXTURE_2D_RW(TYPE,NAME) RWTexture2D<TYPE> NAME
#define TEXTURE_CUBE(TYPE,NAME) SamplerState NAME##Sampler;TextureCube<TYPE> NAME

#define gxSizeOf(T,S) {uint __tempW__,__tempH__;T.GetDimensions(__tempW__,__tempH__);S=int2(__tempW__,__tempH__);}
#define gxSizeOfLevel(T,L,S) {uint __tempW__,__tempH__,__tempN__;T.GetDimensions(L,__tempW__,__tempH__,__tempN__);S=__tempW__;}
#define gxStreamAppend(S,V) (S).Append(V)
#define gxStreamReset(S) (S).RestartStrip()
#define gxSync GroupMemoryBarrierWithGroupSync
#define gxTexture2D(T,C) T.Sample(T##Sampler,C)
#define gxTexture2DArray(T,S,C) T.Sample(T##Sampler,float3(C,S))
#define gxTexture2DArrayLod(T,S,C,L) T.SampleLevel(T##Sampler,float3(C,S),L)
#define gxTexture2DGet(T,C) T[C]
#define gxTexture2DLod(T,C,L) T.SampleLevel(T##Sampler,C,L)
#define gxTexture2DSet(T,C,X) (T[C]=X)
#define gxTextureCube(T,C) T.Sample(T##Sampler,C)
#define gxTextureCubeLod(T,C,L) T.SampleLevel(T##Sampler,C,L)
#define gxTextures2D(T,N,C) T[N].Sample(T##Sampler,C)
#define gxTextures2DArray(T,N,S,C) T[N].Sample(T##Sampler,float3(C,S))
#define gxTextures2DArrayLod(T,N,S,C,L) T[N].SampleLevel(T##Sampler,float3(C,S),L)
#define gxTextures2DLod(T,N,C,L) T[N].SampleLevel(T##Sampler,C,L)
#define gxTexturesCube(T,N,C) T[N].Sample(T##Sampler,C)
#define gxTexturesCubeLod(T,N,C,L) T[N].SampleLevel(T##Sampler,C,L)

#endif

// -------------------------------------------------------------------------------------------------
// Structures
// -------------------------------------------------------------------------------------------------

#ifdef GX_HLSL_DX9

struct FACE
{
  float front : VFACE;

};

float gxFace(FACE face)
{
  return face.front < 0 ? -1 : +1;
}

#else // ------------------------------------------------------------------------------------------

struct FACE
{
  uint front : SV_IsFrontFace;
};

float gxFace(FACE face)
{
  return face.front == 0 ? -1 : +1;
}

#endif

#endif
