// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Camera.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__CAMERA
#define TINMAN__CAMERA

#include "../GpuX.hlsl"
#include "CB_Camera_Static.hlsl"
#include "Configuration.hlsl"

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

/*!
 * @brief Computes the logarithmic depth value.
 *
 * @param z The linear Z-coordinate value, between zero and the coordinate of the far clipping
 *          plane.
 *
 * @returns The logarithmic depth value in the range [0..1].
 */
float ComputeDepth(float z);

/*!
 * @brief Augments the given position with the Z-distance along the camera direction.
 *
 * @param locked   Use locked camera direction?
 * @param position The camera-centric vertex position.
 *
 * @returns The camera-centric vertex position (xyz) and the dot product of the vertex position and
 *          the camera direction (w).
 */
float4 ComputeDotDirection(bool locked, float3 position);

/*!
 * @brief Computes the camera-centric vertex position.
 *
 * @param locked        Use locked camera position?
 * @param position      The world-space vertex position.
 * @param positionExtra The precision delta of the world-space vertex position.
 *
 * @returns The camera-centric vertex position.
 */
float3 ComputePosition(bool locked, float3 position, float3 positionExtra);

/*!
 * @brief Computes the logarithmic depth coordinate.
 *
 * The transformation matrix that has produced 'z' and 'w' must be consistent with
 * 'Configuration.NEAR_AT_ZERO'.
 *
 * @param z The Z-coordinate of the transformed vertex position.
 * @param w The W-coordinate of the transformed vertex position.
 *
 * @returns The logarithmic depth coordinate, to be used as replacement for 'z'.
 */
float ComputeZ(float z, float w);

/*!
 * @brief Projects the given vertex position.
 *
 * @param position The camera-centric vertex position.
 *
 * @returns The projected vertex position.
 */
float4 ProjectPosition(float3 position);

/*!
 * @brief Projects the given vertex position (logarithmic depth).
 *
 * @param position The camera-centric vertex position.
 * @param z        Output of linear depth value (input for 'ComputeZ'), for interpolation in pixel
 *                 shader.
 *
 * @returns The projected vertex position.
 */
float4 ProjectPosition(float3 position, out float z);

/*!
 * @brief Projects the given vertex position.
 *
 * @param position The camera-centric vertex position.
 * @param z        Output of unclamped linear depth values (Z- and W-components), for interpolation
 *                 in pixel shader.
 *
 * @returns The projected vertex position (clamped linear depth).
 */
float4 ProjectPositionShadow(float3 position, out float2 z);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

float ComputeDepth(float z)
{
  return log2(1.0 + max(0.0, z)) / g_viewProjectionFar;
}

float4 ComputeDotDirection(bool locked, float3 position)
{
  return float4(position, dot(position, g_camera[locked ? 5 : 2]));
}

float3 ComputePosition(bool locked, float3 position, float3 positionExtra)
{
  int i;
  float3 high;
  float3 low;

  i = locked ? 3 : 0;
  high = position - g_camera[i];
  low = positionExtra - g_camera[i + 1];

  return high + low;
}

float ComputeZ(float z, float w)
{
  // Map z to [0 .. +w], if necessary.

  if (!NEAR_AT_ZERO)
    z = (z + w) / 2.0;

  // Compute z in the range [0..1].

  z = ComputeDepth(z);

  // Map z to [-1 .. +1], if necessary.

  if (!NEAR_AT_ZERO)
    z = z * 2.0 - 1.0;

  // Map z to [-w .. +w], if necessary.

  return z * w;
}

float4 ProjectPosition(float3 position)
{
  float4 result;

  result = mul(g_viewProjectionMatrix[0], float4(position, 1));
  result.z = ComputeZ(result.z, result.w);

  return result;
}

float4 ProjectPosition(float3 position, out float z)
{
  float4 result;

  result = mul(g_viewProjectionMatrix[0], float4(position, 1));
  result.z = ComputeZ(z = result.z, result.w);

  return result;
}

float4 ProjectPositionShadow(float3 position, out float2 z)
{
  float4 result;

  result = mul(g_viewProjectionMatrix[0], float4(position, 1));
  z = result.zw;

  // Clamp at near-plane to avoid clipping of geometry (early-Z).

  result.z = max(result.z, NEAR_AT_ZERO ? 0.0 : -result.w);

  return result;
}

#endif
