// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Model.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__MODEL__MODEL
#define TINMAN__MODEL__MODEL

#include "../../GpuX.hlsl"
#include "../Camera.hlsl"
#include "../CB_Camera_Static.hlsl"
#include "../CB_Lighting_Dynamic.hlsl"
#include "../CB_Lighting_Static.hlsl"
#include "../Configuration.hlsl"
#include "../Lighting.hlsl"
#include "../PS_Out_TargetDepthLess.hlsl"
#include "../Shadow.hlsl"
#include "../Util.hlsl"
#include "CB_Material_Dynamic1.hlsl"
#include "CB_Material_Dynamic2.hlsl"
#include "CB_Material_Dynamic3.hlsl"
#include "CB_Material_Static.hlsl"
#include "CB_Model_Dynamic.hlsl"
#include "CB_Model_Volatile.hlsl"
#include "SR_Model.hlsl"
#include "VS_Out.hlsl"

// -------------------------------------------------------------------------------------------------
// Constants
// -------------------------------------------------------------------------------------------------

static const int MaterialBaseColor = 0;

static const int MaterialDisplacement = 11;

static const int MaterialEmissive = 1;

static const int MaterialGlossinessOrRoughness = 2;

static const int MaterialMetalness = 3;

static const int MaterialNormal = 4;

static const int MaterialOcclusion = 5;

static const int MaterialOpacity = 6;

static const int MaterialReflectivity = 7;

static const int MaterialSpecularity = 8;

static const int MaterialTranslucency = 9;

static const int MaterialTransmission = 10;

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

#ifdef GX_SHADER_PS
/*!
 * @brief Applies the linear opacity function to the given opacity value.
 *
 * @param alpha  The input opacity value, which is the result of the combination of vertex opacity
 *               and base color opacity.
 * @param coords The texture coordinate sets.
 *
 * @returns The output opacity value.
 */
float MaterialAlpha(float alpha, float4 coords);
#endif

/*!
 * @brief Computes the factor for fading out the 3D model.
 *
 * @param distance The camera distance.
 *
 * @returns The fade factor (0 := fully faded out, 1 := fully faded in).
 */
float ModelFade(float distance);

#ifdef GX_SHADER_PS
PS_Out_TargetDepthLess ModelPixel(VS_Out data, FACE face);
#endif

/*!
 * @brief Computes the reduced world-space position for the given transformed model-space position.
 *
 * @param locked        Use locked camera position instead of actual one?
 * @param position      The transformed model-space position.
 * @param positionExtra The transformed model-space position delta.
 *
 * @returns The position in reduced world-space, i.e. relative to camera position.
 */
float3 ModelPosition(bool locked, float3 position, float3 positionExtra);

/*!
 * @brief Transforms the given model-space position.
 *
 * @param position      The model-space position.
 * @param positionExtra The model-space position delta.
 */
void ModelTransform(inout float3 position, inout float3 positionExtra);

/*!
 * @brief Returns the texture coordinates for the given material texture.
 *
 * @param which  The material texture index.
 * @param coords The texture coordinate sets.
 *
 * @returns The texture coordinates to use for the material texture.
 */
float2 Texture(int which, float4 coords);

/*!
 * @brief Samples the given material texture, returning the selected channel as a scalar.
 *
 * @param which  The material texture index.
 * @param coords The texture coordinate sets.
 *
 * @returns The sampled scalar value.
 */
float Texture_I(int which, float4 coords);

/*!
 * @brief Samples the given material texture, returning the red, green and blue channels as a vector.
 *
 * @param which  The material texture index.
 * @param coords The texture coordinate sets.
 *
 * @returns The sampled color vector.
 */
float3 Texture_RGB(int which, float4 coords);

/*!
 * @brief Samples the given material texture, returning the red, green, blue and alpha channels as a
 *        vector.
 *
 * @param which  The material texture index.
 * @param coords The texture coordinate sets.
 *
 * @returns The sampled color vector.
 */
float4 Texture_RGBA(int which, float4 coords);

/*!
 * @brief Extracts the vertex opacity from the given vertex color tuple.
 *
 * @param color The vertex color tuple (vertex shader input).
 *
 * @returns The vertex alpha value.
 */
float VertexAlpha(float4 color);

/*!
 * @brief Extracts the vertex opacity from the given vertex color tuple.
 *
 * @param color The vertex color tuple (vertex shader input).
 *
 * @returns The vertex alpha value.
 */
float3 VertexColor(float4 color);

/*!
 * @brief Extracts the vertex color, opacity and ambient occlusion from the given vertex color tuple.
 *
 * @param color     The vertex color.
 * @param baseColor The output base color, with opacity.
 * @param occlusion The output ambient occlusion.
 */
void VertexColorAlphaOcclusion(float4 color, out float4 baseColor, out float occlusion);

/*!
 * @brief Extracts the vertex ambient occlusion from the given vertex color tuple.
 *
 * @param color The vertex color tuple (vertex shader input).
 *
 * @returns The vertex ambient occlusion.
 */
float VertexOcclusion(float4 color);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

#ifdef GX_SHADER_PS
float MaterialAlpha(float alpha, float4 coords)
{
  float range;

  // Ignore input alpha, render opaque?

  if (g_materialOpacity > 1.0)
    return 1.0;

  // Modulate with opacity texture map.

  alpha *= Texture_I(MaterialOpacity, coords);

  // Use input alpha, without cutoff?

  if (g_materialOpacity == 0.0)
    return alpha;

  // Apply alpha cutoff, using partial derivatives to create smooth edges.

  range = length(float2(ddx(alpha), ddy(alpha)));

  return smoothstep(max(0.0, g_materialOpacity - range), min(1.0, g_materialOpacity + range), alpha);
}
#endif

float ModelFade(float distance)
{
  return saturate(1.0 - (g_modelFade.x - distance) * g_modelFade.z) * saturate(1.0 - (distance -
      g_modelFade.y) * g_modelFade.w);
}

#ifdef GX_SHADER_PS
PS_Out_TargetDepthLess ModelPixel(VS_Out data, FACE face)
{
  PS_Out_TargetDepthLess o;
  float4 baseColor;
  float opacity;
  float occlusion;

  // Apply distance fade range, base color and opacity.

  {
    float a;

    a = ModelFade(data.DistanceOcclusion.x);

    clip(a - 0.002);

    baseColor = data.Color * g_materialBaseColor * Texture_RGBA(MaterialBaseColor, data.Coords);
    opacity = MaterialAlpha(baseColor.a, data.Coords);
    baseColor.a = a * opacity;
  }

  // Apply line thickness.

  if (THICK_LINES)
  {
    if (g_lineThickness.z > 0.0)
    {
      baseColor.a *= saturate(1.0 - (length(data.Clip) - 1.0) / g_lineThickness.z);
    }
  }

  clip(baseColor.a - 0.002);

  // Compute logarithmic depth value after applying depth bias.

  o.Depth = ComputeDepth(data.Plain.w - data.Plain.w * g_depthBias);

  // Compute normal vector (tangent Z-axis). If necessary, replace with eye-vector to avoid
  // artefacts.
  // => glTF/normalTexture

  {
    float3 plain;
    float sign;

    data.Eye = normalize(data.Eye);
    plain = data.Plain.xyz;

    if (g_tangentMask.z == 0.0)
      data.Normal = cross(ddy(plain), ddx(plain));

    data.Normal = Normalize(data.Normal, data.Eye);
    sign = -gxFace(face);

    // If there is a normal map, tangent-space is required to adjust the normal vector.

    if (g_materialNormal > 0.0)
    {
      // Compute tangent (tangent X-axis) vector. If necessary, replace with dummy to avoid
      // artefacts.

      if (g_tangentMask.x == 0.0)
      {
        float2 uv;
        float2 dx;
        float2 dy;
        float3 px;
        float3 py;

        uv = Texture(MaterialNormal, data.Coords);
        dx = ddx(uv);
        dy = ddy(uv);

        if (length(dx) * length(dy) <= 1E-10)
        {
          dx = float2(1.0, 0.0);
          dy = float2(0.0, 1.0);
        }

        px = ddx(plain);
        py = ddy(plain);
        data.Tangent = (px * dy.y - py * dx.y) / (dx.x * dy.y - dy.x * dx.y);
        data.Tangent = data.Tangent - data.Normal * dot(data.Normal, data.Tangent);
      }

      data.Tangent = Normalize(data.Tangent, float3(1, 0, 0));

      // Compute bitangent (tangent Y-axis) vector. If necessary, replace with dummy to avoid
      // artefacts.

      if (g_tangentMask.y == 0.0)
        data.Bitangent = cross(data.Normal, data.Tangent);

      data.Bitangent = Normalize(data.Bitangent, float3(0, 1, 0));

      // Compute final normal vector.

      data.Normal *= sign;
      data.Tangent *= sign;
      data.Bitangent *= sign;
      data.Normal = normalize(mul((Texture_RGB(MaterialNormal, data.Coords) - 0.5) * float3(
          g_materialNormal, g_materialNormal, 1), float3x3(data.Tangent, data.Bitangent, data.Normal
          )));
    }
    else
      data.Normal *= sign;
  }

  // Compute ambient occlusion.

  if (g_materialOcclusion > 0.0)
  {
    occlusion = Texture_I(MaterialOcclusion, data.Coords) * data.DistanceOcclusion.y;
    occlusion = 1.0 + g_materialOcclusion * (occlusion - 1.0);
  }
  else
    occlusion = 1.0;

  occlusion = ComputeAmbientOcclusion(occlusion);

  if (g_lightEnable == 0.0)
  {
    o.Color.rgb = ComputeToneMap(ComputeFog(baseColor.rgb * occlusion * (0.6 + 0.4 * sqrt(abs(dot(
        data.Normal, data.Eye)))), data.Eye, data.DistanceOcclusion.x));
    o.Color.a = baseColor.a;
  }
  else
  {
    // PBR workflow

    {
      float3 reflectivity;
      float intensity;
      float roughness;
      float metalness;
      float specularity;
      float normalDotEye;
      float3 temp1;
      float2 ggx;
      float3 specular;
      float3 diffuse;
      float3 translucent;
      float3 emissive;
      float3 color;
      float transmission;

      reflectivity = g_materialReflectivity * Texture_RGB(MaterialReflectivity, data.Coords);

      if (g_materialMetalness >= 0.0)
      {
        // => glTF/pbrMetallicRoughness
        //    roughnessFactor * metallicRoughnessTexture.g
        //    metallicFactor * metallicRoughnessTexture.b

        roughness = g_materialGlossinessOrRoughness * Texture_I(MaterialGlossinessOrRoughness, data.
            Coords);
        metalness = g_materialMetalness * Texture_I(MaterialMetalness, data.Coords);

        // => KHR_materials_specular
        //    specularColorFactor * specularColorTexture.rgb

        intensity = min(1.0, pow((g_materialRefraction - 1.0) / (g_materialRefraction + 1.0), 2.0));
        reflectivity = lerp(reflectivity * intensity, baseColor.xyz, metalness);
      }
      else
      {
        // => KHR_materials_pbrSpecularGlossiness
        //    glossinessFactor * specularGlossinessTexture.a
        //    specularFactor * specularGlossinessTexture.rgb

        roughness = 1.0 - g_materialGlossinessOrRoughness * Texture_I(MaterialGlossinessOrRoughness,
            data.Coords);
        metalness = max(max(reflectivity.x, reflectivity.y), reflectivity.z);
      }

      // => KHR_materials_specular
      //    specularFactor *  specularTexture.a

      if (g_materialSpecularity < 0.0)
        specularity = 1.0;
      else
      {
        specularity = g_materialSpecularity * Texture_I(MaterialSpecularity, data.Coords);
      }

      roughness = saturate(roughness);
      normalDotEye = dot(data.Normal, data.Eye);
      intensity = 1.0 - roughness;
      temp1 = (reflectivity + (max(reflectivity, intensity) - reflectivity) * pow(1.0 - normalDotEye,
          5.0)) * max(0.0, sign(normalDotEye));
      ggx = gxTexture2D(g_materialLut, float2(abs(normalDotEye), roughness));

      // Specular

      {
        float3 reflected;
        float lod;

        reflected = reflect(-data.Eye, data.Normal);
        reflected = CubemapWarp(reflected, g_horizon);
        lod = sqrt(roughness) * g_mipmap.x;
        specular = EnvironmentSampleSeamless(reflected, lod) * (temp1 * ggx.x + ggx.y) * specularity
            * occlusion;
      }

      // Diffuse (ambient, directional, translucent)

      {
        float3 temp2;
        float factor;
        float shadow;
        float3 ambient;
        float3 directionalSample;
        float3 directional;

        temp2 = temp1 * specularity * ggx.x + ggx.y;
        temp1 = (reflectivity + (-reflectivity + 1.0) / 21.0) * specularity;
        temp1 = temp2 * temp1 / (-temp1 + 1.0 / (1.0 - (ggx.x + ggx.y)));
        factor = dot(data.Normal, g_lightVector);
        shadow = g_lightEnable < 0.0 ? 1.0 : ComputeShadow(data.Shadow);
        ambient = EnvironmentSample(data.Normal, g_mipmap.x) * occlusion * g_lighting.x;
        directionalSample = EnvironmentSample(-g_lightVector, g_mipmap.y);
        directional = directionalSample * saturate(-factor) * g_lighting.y;
        diffuse = (ambient + directional * shadow) * (temp1 + lerp(baseColor.xyz, 0.0, metalness) *
            (temp1 + 1.0 - temp2));

        if (g_materialTranslucency.r >= 0.0)
        {
          translucent = g_materialTranslucency * Texture_RGB(MaterialTranslucency, data.Coords);
          translucent = directionalSample * (temp1 + lerp(translucent, 0.0, metalness) * (temp1 +
              1.0 - temp2)) * saturate(factor) * g_lighting.y;
          diffuse += translucent * shadow;
        }
        else
          translucent = 0.0;
      }

      // Transmission

      transmission = g_materialTransmission;

      if (transmission > 0.0)
      {
        transmission *= g_materialTransmission * Texture_I(MaterialTransmission, data.Coords);
      }

      // Emissive

      if (g_materialEmissive.r < 0.0)
        emissive = 0.0;
      else
      {
        emissive = g_materialEmissive * Texture_RGB(MaterialEmissive, data.Coords);
      }

      // High-dynamic range color

      color = specular + diffuse + emissive;

      // Fog

      color = ComputeFog(color, data.Eye, data.DistanceOcclusion.x);

      // Tone mapping

      color = ComputeToneMap(color);

      // Final color

      o.Color = float4(color, baseColor.a);

      // Debug mode

      if (g_materialDebug != 0)
      {
        float3 debug;

        debug = float3(1, 0, 1);

        if (g_materialDebug == 1)
          debug = baseColor.rgb;

        if (g_materialDebug == 2)
          debug = diffuse;

        if (g_materialDebug == 3)
          debug = emissive;

        if (g_materialDebug == 4)
          debug = FromSrgb(float3(ggx, 0));

        if (g_materialDebug == 5)
          debug = FromSrgb(data.Normal / 2.0 + 0.5);

        if (g_materialDebug == 6)
          debug = FromSrgb(float3(occlusion, roughness, metalness));

        if (g_materialDebug == 7)
          debug = FromSrgb(float3(1.0 - opacity, transmission, 1.0 - specularity));

        if (g_materialDebug == 8)
          debug = reflectivity;

        if (g_materialDebug == 9)
          debug = specular;

        if (g_materialDebug == 10)
        {
          debug = saturate(float3(sign(g_materialNormal) * (1.0 - g_tangentMask.x), sign(
              g_materialNormal) * (1.0 - g_tangentMask.y), 1.0 - g_tangentMask.z));
        }

        if (g_materialDebug == 11)
          debug = FromSrgb(float3(data.Coords.xy, 0));

        if (g_materialDebug == 12)
          debug = FromSrgb(float3(data.Coords.zw, 0));

        if (g_materialDebug == 13)
          debug = translucent;

        o.Color = float4(debug, 1);
      }
    }
  }

  return o;
}
#endif

float3 ModelPosition(bool locked, float3 position, float3 positionExtra)
{
  position += g_modelCenter[0] - g_camera[locked ? 3 : 0];
  positionExtra += g_modelCenter[1] - g_camera[locked ? 4 : 1];

  return position + positionExtra;
}

void ModelTransform(inout float3 position, inout float3 positionExtra)
{
  position = mul(g_modelMatrix, float4(position, 1));
  positionExtra = mul(g_modelMatrixIT, float4(positionExtra, 0));
}

float2 Texture(int which, float4 coords)
{
  return g_materialInfo[which].x == 0.0 ? coords.xy : coords.zw;
}

float Texture_I(int which, float4 coords)
{
  return gxTextures2D(g_material, which, Texture(which, coords))[g_materialInfo[which].y];
}

float3 Texture_RGB(int which, float4 coords)
{
  return gxTextures2D(g_material, which, Texture(which, coords)).rgb;
}

float4 Texture_RGBA(int which, float4 coords)
{
  return gxTextures2D(g_material, which, Texture(which, coords));
}

float VertexAlpha(float4 color)
{
  return lerp(1.0, color.a, saturate(g_vertexColor.y));
}

float3 VertexColor(float4 color)
{
  return lerp(1.0, FromSrgb(color.rgb), g_vertexColor.x);
}

void VertexColorAlphaOcclusion(float4 color, out float4 baseColor, out float occlusion)
{
  baseColor = float4(VertexColor(color), VertexAlpha(color));
  occlusion = VertexOcclusion(color);
}

float VertexOcclusion(float4 color)
{
  return lerp(1.0, color.a, saturate(-g_vertexColor.y));
}

#endif
