// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Terrain.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__TERRAIN__TERRAIN
#define TINMAN__TERRAIN__TERRAIN

#include "../../GpuX.hlsl"
#include "../Camera.hlsl"
#include "../CB_Lighting_Dynamic.hlsl"
#include "../CB_Lighting_Static.hlsl"
#include "../Configuration.hlsl"
#include "../Lighting.hlsl"
#include "../Util.hlsl"
#include "CB_Material_Dynamic.hlsl"
#include "CB_Material_Static.hlsl"
#include "CB_Material_Volatile1.hlsl"
#include "CB_Material_Volatile2.hlsl"
#include "CB_Terrain_Dynamic.hlsl"
#include "CB_Terrain_Static.hlsl"
#include "CB_Texture_Static.hlsl"
#include "CB_Texture_Volatile.hlsl"
#include "Displacement.hlsl"
#include "DisplacementFixed.hlsl"
#include "DisplacementVarying.hlsl"
#include "GS_In_Material.hlsl"
#include "Material.hlsl"
#include "PS_In_Material.hlsl"
#include "SR_Terrain.hlsl"
#include "VS_In.hlsl"

// -------------------------------------------------------------------------------------------------
// Constants
// -------------------------------------------------------------------------------------------------

static const float2 WireframeAlpha = float2(0.9, 0.1);

static const float3 WireframeDark = float3(0, 0, 0);

static const float3 WireframeLight = float3(1, 1, 1);

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

/*!
 * @brief Computes the resulting light for the given terrain material properties.
 *
 * @param albedo       The albedo color (linear sRGB).
 * @param occlusion    Scale factor for ambient occlusion, in the range [0..1].
 * @param reflectivity The material reflectivity (linear sRGB).
 * @param gloss        The material glossiness.
 * @param eye          The unit-length eye vector.
 * @param normal       The unit-length normal vector.
 * @param distance     The camera distance.
 * @param shadow       The shadow factor: 0 = full shadow, 1 = no shadow.
 *
 * @returns The resulting color.
 */
float3 ComputeColor(float3 albedo, float occlusion, float3 reflectivity, float gloss, float3 eye, 
    float3 normal, float distance, float shadow);

Displacement ComputeDisplacement(float3 position, float3 positionDelta, float3 normal, float3
    normalBase, int4 materialIds, float4 materialWeights, float2 levelOfDetail, bool computeDelta);

Displacement ComputeDisplacement(VS_In vertex);

DisplacementFixed ComputeDisplacementFixed(float3 position, float3 positionDelta, float3 normal, 
    float3 normalBase, int4 materialIds, float2 levelOfDetail);

DisplacementVarying ComputeDisplacementVarying(DisplacementFixed displacement, float4
    materialWeights, bool computeDelta);

float DecodeScalar(float4 samples[TextureCount + 2], float4 mapping);

/*!
 * @brief Fades the given displacement value to zero when reaching the end of the layer blend range.
 *
 * This will avoid Z-fighting artefacts with adjacent, non-displacement terrain layers.
 *
 * @param displacement The displacement value, in terrain-space.
 * @param distance     The distance from the vertex to the camera position.
 *
 * @returns The faded displacement value.
 */
float DisplacementFade(float displacement, float distance);

/*!
 * @brief Computes a material height for use in a vertex / domain / geometry shader.
 *
 * @param id       The material ID.
 * @param coords   The world-space coordinates.
 * @param normal   Unit-length surface normal vector.
 * @param distance Distance of vertex to camera.
 * @param area     Area of vertex neighborhood.
 *
 * @returns The material height (x) and absolute height difference to maximum LOD.
 */
float2 DisplacementMaterial(int id, float3 coords, float3 normal, float distance, float area);

/*!
 * @brief Computes a material height for use in a vertex / domain / geometry / shader.
 *
 * @param materialIds     The material IDs.
 * @param materialWeights The material weights.
 * @param coords          The world-space coordinates.
 * @param normal          Unit-length surface normal vector.
 * @param distance        Distance of vertex to camera.
 * @param area            Area of vertex neighborhood.
 *
 * @returns The material height (x) and absolute height difference to maximum LOD.
 */
float2 DisplacementMaterials(int4 materialIds, float4 materialWeights, float3 coords, float3 normal, 
    float distance, float area);

/*!
 * @brief Applies the terrain highlight.
 *
 * @param color The input color.
 *
 * @returns The highlighted output color.
 */
float4 Highlight(float4 color);

/*!
 * @brief Computes the distance-based layer blend factor.
 *
 * @param distance Distance from fragment to camera.
 *
 * @returns The blend factor.
 */
float LayerBlend(float distance);

#ifdef GX_SHADER_PS
/*!
 * @brief Computes the distance-based layer blend factor.
 *
 * The current fragment is discarded if the resulting blend factory is less than zero.
 *
 * @param distance Distance from fragment to camera.
 *
 * @returns The blend factor.
 */
float LayerBlendClip(float distance);
#endif

/*!
 * @brief Applies material albedo blending (see Tinman.Engine.Components.TerrainMaterialBlend)
 *
 * @param materialAlbedo          The material albedo (linear sRGB).
 * @param materialAlbedoGrayscale Controls saturation of the material albedo.
 * @param materialBlend           Blend factor for material albedo.
 * @param vertexColor             The vertex color (linear sRGB).
 * @param vertexColorBrightness   Controls brightness of the vertex colors.
 * @param vertexColorBlend        Blend factor for vertex color.
 *
 * @returns The resulting material albedo.
 */
float3 MaterialAlbedoBlend(float3 materialAlbedo, float materialAlbedoGrayscale, float materialBlend, 
    float3 vertexColor, float vertexColorBrightness, float vertexColorBlend);

/*!
 * @brief Computes the pixel-shader output color for the current blend mode.
 *
 * @param color The final material color.
 * @param blend The layer blend factor.
 *
 * @returns The pixel-shader output color.
 */
float4 MaterialBlend(float3 color, float blend);

/*!
 * @brief Computes the weight for height-based blending.
 *
 * @param height The displacement height value.
 * @param weight The material weight value.
 *
 * @returns The blend weight.
 */
float MaterialBlendWeight(float height, float weight);

/*!
 * @brief Computes the material color.
 *
 * @param material The terrain material sample.
 * @param eye      Eye vector, in light-space.
 * @param distance Distance from fragment to camera.
 * @param shadow   Shadow factor (0 := full shadow, 1 := no shadow).
 *
 * @returns The resulting color.
 */
float3 MaterialColor(Material material, float3 eye, float distance, float shadow);

PS_In_Material MaterialCompute(GS_In_Material material);

PS_In_Material MaterialCompute(GS_In_Material material, DisplacementVarying displacement);

/*!
 * @brief Constructs texture coordinates for tri-planar texture mapping.
 *
 * The material texture coordinate scale factors of g_materialSettings are applied.
 *
 * @param n        The material index.
 * @param position The vertex position.
 * @param normal   The unit-length vertex normal vector.
 * @param x        Texture coordinates for the YZ-plane.
 * @param y        Texture coordinates for the XZ-plane.
 * @param z        Texture coordinates for the XY-plane.
 */
void MaterialComputeCoords(int n, float3 position, float3 normal, out float2 x, out float2 y, out 
    float2 z);

/*!
 * @brief Computes material weights.
 *
 * The components of 'tm' and 'vw' vw are tied together.
 *
 * @param vm Holds an input material ID in each component.
 * @param vw Holds an input material weight in each component.
 * @param tm Holds an output material ID in each component.
 *
 * @returns The computed material weights, corresponding to the components in 'tm'.
 */
float4 MaterialComputeWeights(float4 vm, float4 vw, float4 tm);

#ifdef GX_SHADER_PS
/*!
 * @brief Adjusts the given color in order to generate coordinate system grid lines.
 *
 * @param coordinate The coordinate from which to generate a grid line.
 * @param color      The input color.
 * @param grid       The grid color.
 *
 * @returns The resulting color.
 */
float3 MaterialGridLine(float coordinate, float3 color, float3 grid);
#endif

/*!
 * @brief Mixes two input vertices evenly.
 *
 * @param a First input vertex to mix.
 * @param b Second input vertex to mix.
 *
 * @returns The mixed vertex.
 */
GS_In_Material MaterialMix2(GS_In_Material a, GS_In_Material b);

/*!
 * @brief Mixes three input vertices.
 *
 * @param a First input vertex to mix.
 * @param b Second input vertex to mix.
 * @param c Third input vertex to mix.
 * @param f The mix factors.
 *
 * @returns The mixed vertex.
 */
GS_In_Material MaterialMix3(GS_In_Material a, GS_In_Material b, GS_In_Material c, float3 f);

/*!
 * @brief Returns the rotation matrix for material coordinates.
 *
 * @returns The matrix.
 */
float2x2 MaterialRotation();

/*!
 * @brief Computes a material sample for use in a pixel shader.
 *
 * @param n      The material index.
 * @param coords The world-space coordinates.
 * @param normal Unit-length surface normal vector.
 * @param color  The vertex color.
 *
 * @returns The material sample.
 */
Material MaterialSample(int n, float3 coords, float3 normal, float3 color);

/*!
 * @brief Computes the tri-planar tangent space.
 *
 * @param normal   The normal vector.
 * @param tangentX The tangent vector for the X-axis.
 * @param tangentY The tangent vector for the Y-axis.
 * @param tangentZ The tangent vector for the Z-axis.
 */
void MaterialTangentSpace(float3 normal, out float3 tangentX, out float3 tangentY, out float3
    tangentZ);

/*!
 * @brief Computes the tessellation factor for the given size in terrain-space.
 *
 * @param size     The size in terrain-space.
 * @param distance Distance to camera.
 *
 * @returns The tessellation edge factor, representing the number of resulting line segments per
 *          input edge:
 *          1.0 := no tessellation (one edge -> one line segment)
 *          2.0 := one level of recursive subdivision (one edge -> two line segments)
 *          4.0 := two levels of recursive subdivision (one edge -> four line segments)
 */
float MaterialTessellationFactor(float size, float distance);

float MaterialTessellationFactor(GS_In_Material a0, GS_In_Material a1, DisplacementVarying b0, 
    DisplacementVarying b1);

/*!
 * @brief Computes the blend factor of a texture atlas tile.
 *
 * @param atlas    The texture atlas.
 * @param coords   Texture coordinates for terrain (see TextureAtlasCoords).
 * @param distance Distance between camera and vertex.
 *
 * @returns The blend factor.
 */
float TextureBlend(int atlas, float2 coords, float distance);

/*!
 * @brief Clamps the given texture coordinates.
 *
 * @param coordinates The texture coordinates to clamp.
 * @param bounds      The coordinate bounds: Tinman.Terrain.Rendering.TextureTile.Bounds
 *
 * @returns The clamped texture coordinates.
 */
float2 TextureClamp(float2 coordinates, float4 bounds);

/*!
 * @brief Computes atlas texture coordinates.
 *
 * @param atlas          The texture atlas, one of: AOR, NOR
 * @param texCoords      Global terrain mesh coordinates:
 *
 *                       Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.FloatHigh)
 *                       or
 *
 *                       Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.Float)
 * @param texCoordsExtra Global terrain mesh coordinates:
 *
 *                       Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.FloatLow)
 *                       or zero
 * @param coordsBase     Output coordinates for base texture.
 * @param coords         Output coordinates for texture.
 */
void TextureCoords(int atlas, float3 texCoords, float3 texCoordsExtra, out float2 coordsBase, out 
    float2 coords);

/*!
 * @brief Computes atlas texture coordinates.
 *
 * @param texCoordsHigh Global terrain mesh coordinates:
 *
 *                      Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.FloatHigh)
 *                      or
 *
 *                      Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.Float)
 * @param texCoordsLow  Global terrain mesh coordinates:
 *
 *                      Tinman.Terrain.Vertices.Formats.HeightmapCoordinates(CoordinatePrecision.FloatLow)
 *                      or zero
 * @param offset        Texturing coefficient: Tinman.Terrain.Rendering.TextureTile.Offset
 * @param u             Texturing coefficient: Tinman.Terrain.Rendering.TextureTile.U
 * @param v             Texturing coefficient: Tinman.Terrain.Rendering.TextureTile.V
 *
 * @returns The texture coordinates.
 */
float2 TextureCoords(float3 texCoordsHigh, float3 texCoordsLow, float4 offset, float3 u, float3 v);

/*!
 * @brief Fetches a material sample from the texture atlases.
 *
 * @param vertex   The default texture sample to use.
 * @param coords   The texture coordinates for BASE and TILE.
 * @param distance The distance between the camera and the vertex.
 * @param opacity  The output opacity.
 *
 * @returns The material sample.
 */
Material TextureSample(float4 vertex, float4 coords[TextureCount], float distance, out float opacity
);

/*!
 * @brief Computes the per-vertex terrain LOD value.
 *
 * @param levelOfDetail The LOD coefficients of the vertex, see 'Semantic.Area'.
 * @param distanceSqr   Squared distance from the camera position to the vertex.
 *
 * @returns The LOD value (0 := base LOD, 1 := maximum LOD).
 */
float VertexBlend(float2 levelOfDetail, float distanceSqr);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

float3 ComputeColor(float3 albedo, float occlusion, float3 reflectivity, float gloss, float3 eye, 
    float3 normal, float distance, float shadow)
{
  float normalDotEye;
  float3 result;

  normalDotEye = dot(normal, eye);

  // All shader invocations in the same GPU wavefront will take the same path here.

  if (g_lightEnable == 0.0)
  {
    // Plain albedo with baked ambient occlusion.
    // Modulate with some fake lighting.

    result = ComputeFog(albedo * ComputeAmbientOcclusion(occlusion) * (0.6 + 0.4 * sqrt(abs(
        normalDotEye))), eye, distance);
  }
  else
  {
    float roughness;
    float metalness;
    float3 temp1;
    float2 ggx;
    float3 specular;
    float3 diffuse;
    float3 color;

    occlusion = ComputeAmbientOcclusion(occlusion);
    roughness = 1.0 - gloss;
    metalness = max(max(reflectivity.x, reflectivity.y), reflectivity.z);
    temp1 = (reflectivity + (max(reflectivity, gloss) - reflectivity) * pow(1.0 - normalDotEye, 5.0)
        ) * max(0.0, sign(normalDotEye));
    ggx = ComputeFakeGgx(abs(normalDotEye), roughness);

    // Specular

    {
      float3 reflected;
      float lod;

      reflected = reflect(-eye, normal);
      reflected = CubemapWarp(reflected, g_horizon);
      lod = sqrt(roughness) * g_mipmap.x;
      specular = EnvironmentSampleSeamless(reflected, lod) * (temp1 * ggx.x + ggx.y) * occlusion;
    }

    // Diffuse (ambient, directional)

    {
      float3 temp2;
      float factor;
      float3 ambient;
      float3 directional;

      temp2 = temp1 * ggx.x + ggx.y;
      temp1 = reflectivity + (-reflectivity + 1.0) / 21.0;
      temp1 = temp2 * temp1 / (-temp1 + 1.0 / (1.0 - (ggx.x + ggx.y)));
      factor = dot(normal, g_lightVector);
      ambient = EnvironmentSample(normal, g_mipmap.x) * occlusion * g_lighting.x;
      directional = EnvironmentSample(-g_lightVector, g_mipmap.y) * saturate(-factor) * g_lighting.y;
      diffuse = (ambient + directional * shadow) * (temp1 + lerp(albedo, 0.0, metalness) * (temp1 +
          1.0 - temp2));
    }

    // High-dynamic range color

    color = specular + diffuse;

    // Fog

    color = ComputeFog(color, eye, distance);

    // Tone mapping

    result = ComputeToneMap(color);
  }

  return result;
}

Displacement ComputeDisplacement(float3 position, float3 positionDelta, float3 normal, float3
    normalBase, int4 materialIds, float4 materialWeights, float2 levelOfDetail, bool computeDelta)
{
  Displacement displacement;

  displacement.Fixed = ComputeDisplacementFixed(position, positionDelta, normal, normalBase,
      materialIds, levelOfDetail);
  displacement.Varying = ComputeDisplacementVarying(displacement.Fixed, materialWeights,
      computeDelta);

  return displacement;
}

Displacement ComputeDisplacement(VS_In vertex)
{
  return ComputeDisplacement(vertex.Position, vertex.PositionDelta, DecodeShift(vertex.Normal),
      DecodeShift(vertex.NormalBase), vertex.MaterialIds, vertex.MaterialWeights, vertex.
      LevelOfDetail, false);
}

DisplacementFixed ComputeDisplacementFixed(float3 position, float3 positionDelta, float3 normal, 
    float3 normalBase, int4 materialIds, float2 levelOfDetail)
{
  DisplacementFixed displacement;
  float distanceSqr;
  float2x2 rotation;

  displacement.Position = ComputePosition(false, position, positionDelta);
  displacement.Shadow = ComputePosition(true, position, positionDelta);
  distanceSqr = dot(displacement.Shadow, displacement.Shadow);
  normal = normalize(lerp(normalBase, normal, VertexBlend(levelOfDetail, distanceSqr)));
  displacement.Distance = sqrt(distanceSqr);
  displacement.Normal = mul(g_lightSpace, normal);
  displacement.Eye = mul(g_lightSpace, -displacement.Shadow / displacement.Distance);
  displacement.Area = levelOfDetail.y;
  displacement.MaterialIds = materialIds;

  // Compute initial material coordinates.

  displacement.MaterialCoords = mul(g_lightSpace, position - g_materialOrigin[0] + (positionDelta -
      g_materialOrigin[1]));

  if (g_materialTransform.z > 0.0)
  {
    float horizontal;
    float vertical;
    float angle;
    float scale;

    // Compute projected material coordinates.

    horizontal = length(displacement.MaterialCoords.xz);
    vertical = g_materialTransform.z + displacement.MaterialCoords.y;
    angle = atan2(horizontal, vertical);
    scale = horizontal > 1E-10 ? angle * g_materialTransform.z / horizontal : 0.0;
    displacement.MaterialCoords.y = length(float3(displacement.MaterialCoords.x, vertical,
        displacement.MaterialCoords.z)) - g_materialTransform.z;
    displacement.MaterialCoords.xz *= scale;
  }

  // Compute adjusted material coordinates.

  rotation = MaterialRotation();
  displacement.MaterialCoords.xz = mul(rotation, displacement.MaterialCoords.xz);
  displacement.MaterialCoords += g_materialTranslate;
  displacement.MaterialNormal.y = displacement.Normal.y;
  displacement.MaterialNormal.xz = mul(rotation, displacement.Normal.xz);

  return displacement;
}

DisplacementVarying ComputeDisplacementVarying(DisplacementFixed displacement, float4
    materialWeights, bool computeDelta)
{
  DisplacementVarying result;
  float3 temp;

  if (TERRAIN_MATERIAL_DISPLACEMENT)
  {
    float scale;

    scale = DisplacementFade(g_displacementSettings.x, displacement.Distance);
    result.Displacement = float2(scale, scale);

    if (result.Displacement.x > 0.0)
    {
      result.Displacement *= DisplacementMaterials(displacement.MaterialIds, materialWeights,
          displacement.MaterialCoords, displacement.MaterialNormal, displacement.Distance,
          computeDelta ? displacement.Area : 0.0);
    }
  }
  else
    result.Displacement = 0.0;

  temp = mul(displacement.Normal, g_lightSpace) * result.Displacement.x;
  result.Position = displacement.Position + temp;
  result.Shadow = ComputeDotDirection(true, displacement.Shadow + temp);

  return result;
}

float DecodeScalar(float4 samples[TextureCount + 2], float4 mapping)
{
  float result;

  // Fetch texture atlas sample value.

  result = samples[mapping.x][mapping.y];

  // Apply linear function.

  result = result * mapping.z + mapping.w;

  return result;
}

float DisplacementFade(float displacement, float distance)
{
  return displacement * saturate(distance * g_layerBlend.z + g_layerBlend.w);
}

float2 DisplacementMaterial(int id, float3 coords, float3 normal, float distance, float area)
{
  if (TERRAIN_MATERIAL_DISPLACEMENT)
  {
    float4 m;
    float4 d;
    float sliceY;
    float sliceX;
    float3 f;
    float2 planarX;
    float2 planarY;
    float2 planarZ;
    float size;
    float2 level;
    float height;
    float delta;

    m = g_materialSettings[id];
    d = g_displacementInfo;
    sliceX = id - d.x;
    sliceY = id - d.y;

    // Setup tri-planar texturing coefficients.

    f = (m.y == 0.0 ? float3(0, 1, 0) : normal * normal) * float3(m.w, m.z, m.w);

    MaterialComputeCoords(id, coords, normal, planarX, planarY, planarZ);

    // Compute mipmap level for one screen pixel and fetch displacement sample.

    size = distance / g_vertexBlend.x;
    level = max(0.0, log2(d.zw * size * m.xy));
    height = 0.0 + f.x * (gxTextures2DArrayLod(g_displacement, 1, sliceY, planarX, level.y).r - 0.5)
        + f.y * (gxTextures2DArrayLod(g_displacement, 0, sliceX, planarY, level.x).r - 0.5) + f.z *
        (gxTextures2DArrayLod(g_displacement, 1, sliceY, planarZ, level.y).r - 0.5);

    if (area > 0.0)
    {
      // Compute mipmap level for area of vertex neighborhood and fetch displacement sample.

      size = sqrt(area);
      level = max(0.0, log2(d.zw * size * m.xy));
      delta = 0.0 + f.x * (gxTextures2DArrayLod(g_displacement, 1, sliceY, planarX, level.y).r - 0.5
          ) + f.y * (gxTextures2DArrayLod(g_displacement, 0, sliceX, planarY, level.x).r - 0.5) + f.
          z * (gxTextures2DArrayLod(g_displacement, 1, sliceY, planarZ, level.y).r - 0.5);
      delta = abs(delta - height);
    }
    else
      delta = 0.0;

    return float2(height, delta);
  }

  return 0.0;
}

float2 DisplacementMaterials(int4 materialIds, float4 materialWeights, float3 coords, float3 normal, 
    float distance, float area)
{
  float2 result;

  result = 0.0;

  if (materialWeights.r > 0.0)
  {
    result += DisplacementMaterial(materialIds.r, coords, normal, distance, area) * materialWeights.
        r;
  }

  if (materialWeights.g > 0.0)
  {
    result += DisplacementMaterial(materialIds.g, coords, normal, distance, area) * materialWeights.
        g;
  }

  if (materialWeights.b > 0.0)
  {
    result += DisplacementMaterial(materialIds.b, coords, normal, distance, area) * materialWeights.
        b;
  }

  if (materialWeights.a > 0.0)
  {
    result += DisplacementMaterial(materialIds.a, coords, normal, distance, area) * materialWeights.
        a;
  }

  return result;
}

float4 Highlight(float4 color)
{
  float alpha;

  alpha = max(color.a, g_terrainHighlight.a);

  return float4(lerp(color.rgb, g_terrainHighlight.rgb / alpha, g_terrainHighlight.a), alpha);
}

float LayerBlend(float distance)
{
  return distance * g_layerBlend.x + g_layerBlend.y;
}

#ifdef GX_SHADER_PS
float LayerBlendClip(float distance)
{
  float f;

  clip(f = LayerBlend(distance));

  return f;
}
#endif

float3 MaterialAlbedoBlend(float3 materialAlbedo, float materialAlbedoGrayscale, float materialBlend, 
    float3 vertexColor, float vertexColorBrightness, float vertexColorBlend)
{
  float3 a;
  float3 b;
  float m;

  a = lerp(materialAlbedo, dot(materialAlbedo, Luminance), materialAlbedoGrayscale);
  m = max(max(vertexColor.r, vertexColor.g), vertexColor.b);
  b = m == 0.0 ? vertexColor : lerp(vertexColor, vertexColor / m, vertexColorBrightness);

  return a * b * (1.0 - materialBlend - vertexColorBlend) + a * materialBlend + b * vertexColorBlend;
}

float4 MaterialBlend(float3 color, float blend)
{
  return float4(Highlight(float4(color, 1)).xyz, blend);
}

float MaterialBlendWeight(float height, float weight)
{
  if (TERRAIN_MATERIAL_DISPLACEMENT)
  {
    if (weight < 1.0 / 255.0)
      return 0.0;

    weight *= pow(1.0 + (abs(height) + g_materialHeightBlend.x) * g_materialHeightBlend.y,
        g_materialHeightBlend.z);
  }

  return weight;
}

float3 MaterialColor(Material material, float3 eye, float distance, float shadow)
{
  return ComputeColor(material.Albedo, material.Occlusion, material.Reflectivity, material.Gloss,
      eye, material.Normal, distance, shadow);
}

PS_In_Material MaterialCompute(GS_In_Material material)
{
  return MaterialCompute(material, ComputeDisplacementVarying(material.Fixed, material.WeightsA,
      false));
}

PS_In_Material MaterialCompute(GS_In_Material material, DisplacementVarying displacement)
{
  PS_In_Material result;

  result.Distance = material.Fixed.Distance;
  result.Normal = material.Fixed.Normal;
  result.Eye = material.Fixed.Eye;
  result.MaterialCoords = material.Fixed.MaterialCoords;
  result.MaterialNormal = material.Fixed.MaterialNormal;
  result.Occlusion = material.Color.a;
  result.Shadow = displacement.Shadow;
  result.Color = material.Color.rgb;
  result.Position = ProjectPosition(displacement.Position);
  result.MaterialWeights = material.WeightsB;

  return result;
}

void MaterialComputeCoords(int n, float3 position, float3 normal, out float2 x, out float2 y, out 
    float2 z)
{
  float px;
  float py;
  float pz;

  px = position.x * g_materialSettings[n].y;
  py = -position.y * g_materialSettings[n].y;
  pz = -position.z * g_materialSettings[n].y;
  x = normal.x < 0.0 ? float2(0.5 - pz, py) : float2(pz, py);
  z = normal.z < 0.0 ? float2(0.5 - px, py) : float2(px, py);
  px = position.x * g_materialSettings[n].x;
  pz = position.z * g_materialSettings[n].x;
  y = normal.y < 0.0 ? float2(0.5 - pz, 0.5 - px) : float2(px, pz);
}

float4 MaterialComputeWeights(float4 vm, float4 vw, float4 tm)
{
  return float4(dot(-abs(sign(-vm + tm.x)) + 1.0, vw), dot(-abs(sign(-vm + tm.y)) + 1.0, vw), dot(-
      abs(sign(-vm + tm.z)) + 1.0, vw), dot(-abs(sign(-vm + tm.w)) + 1.0, vw));
}

#ifdef GX_SHADER_PS
float3 MaterialGridLine(float coordinate, float3 color, float3 grid)
{
  float ddxy;
  float temp;

  ddxy = sqrt(Sqr(ddx(coordinate)) + Sqr(ddy(coordinate)));
  temp = coordinate - floor(coordinate / g_materialGridLines) * g_materialGridLines;

  return lerp(color, grid, saturate(smoothstep(ddxy * 2.0, ddxy / 2.0, abs(temp -
      g_materialGridLines / 2.0))));
}
#endif

GS_In_Material MaterialMix2(GS_In_Material a, GS_In_Material b)
{
  GS_In_Material c;

  // Interpolate unit-length vectors.

  c.Fixed.Eye = normalize(a.Fixed.Eye + b.Fixed.Eye);
  c.Fixed.Normal = normalize(a.Fixed.Normal + b.Fixed.Normal);
  c.Fixed.MaterialNormal = normalize(a.Fixed.MaterialNormal + b.Fixed.MaterialNormal);

  // Interpolate linear values.

  c.Color = (a.Color + b.Color) / 2.0;
  c.Fixed.MaterialCoords = (a.Fixed.MaterialCoords + b.Fixed.MaterialCoords) / 2.0;
  c.Fixed.Distance = (a.Fixed.Distance + b.Fixed.Distance) / 2.0;
  c.Fixed.Position = (a.Fixed.Position + b.Fixed.Position) / 2.0;
  c.Fixed.Shadow = (a.Fixed.Shadow + b.Fixed.Shadow) / 2.0;

  // Interpolate material token.

  if (all(a.Fixed.MaterialIds == b.Fixed.MaterialIds))
  {
    // In most cases, the material IDs will be identical. In this case, just sum up the weights.

    c.Fixed.MaterialIds = a.Fixed.MaterialIds;
    c.WeightsA = (a.WeightsA + b.WeightsA) / 2.0;
  }
  else
  {
    int4 idsA;
    int4 idsB;
    float4 weightsA;
    float4 weightsB;
    float weightSum;
    int countA;
    int countB;

    // Need to mix two material tokens.

    idsA = a.Fixed.MaterialIds;
    idsB = b.Fixed.MaterialIds;
    weightsA = a.WeightsA;
    weightsB = b.WeightsA;
    weightSum = dot((weightsA + weightsB) / 2.0, 1.0);
    countA = (int) dot(sign(weightsA), 1.0);
    countB = (int) dot(sign(weightsB), 1.0);
    c.Fixed.MaterialIds = 0;
    c.WeightsA = 0.0;

    // Two material tokens can contain at most 8 distinct materials.

    while (max(countA, countB) > 0)
    {
      int id;
      float weight;

      if (idsA.x > idsB.x)
      {
        countA--;

        id = idsA.x;
        weight = weightsA.x;
        idsA = int4(idsA.yzw, 0);
        weightsA = float4(weightsA.yzw, 0);
      }
      else if (idsA.x < idsB.x)
      {
        countB--;

        id = idsB.x;
        weight = weightsB.x;
        idsB = int4(idsB.yzw, 0);
        weightsB = float4(weightsB.yzw, 0);
      }
      else
      {
        countA--;
        countB--;

        id = idsA.x;
        weight = weightsA.x + weightsB.x;
        idsA = int4(idsA.yzw, 0);
        weightsA = float4(weightsA.yzw, 0);
        idsB = int4(idsB.yzw, 0);
        weightsB = float4(weightsB.yzw, 0);
      }

      weight *= 0.5;

      if (weight > c.WeightsA.x)
      {
        c.Fixed.MaterialIds = int4(id, c.Fixed.MaterialIds.xyz);
        c.WeightsA = float4(weight, c.WeightsA.xyz);
      }
      else if (weight > c.WeightsA.y)
      {
        c.Fixed.MaterialIds.yzw = int3(id, c.Fixed.MaterialIds.yz);
        c.WeightsA.yzw = float3(weight, c.WeightsA.yz);
      }
      else if (weight > c.WeightsA.z)
      {
        c.Fixed.MaterialIds.zw = int2(id, c.Fixed.MaterialIds.z);
        c.WeightsA.zw = float2(weight, c.WeightsA.z);
      }
      else if (weight > c.WeightsA.w)
      {
        c.Fixed.MaterialIds.w = id;
        c.WeightsA.w = weight;
      }
    }

    // Scale weights in order to retain the weight sum. Otherwise, visual artefacts may appear,
    // for example slightly darkened areas because of weight sums less than one.

    c.WeightsA *= weightSum / dot(c.WeightsA, 1.0);
  }

  c.WeightsB = MaterialComputeWeights(c.Fixed.MaterialIds, c.WeightsA, g_materialIds);

  return c;
}

GS_In_Material MaterialMix3(GS_In_Material a, GS_In_Material b, GS_In_Material c, float3 f)
{
  GS_In_Material d;

  // Interpolate unit-length vectors.

  d.Fixed.Eye = normalize(a.Fixed.Eye * f.x + b.Fixed.Eye * f.y + c.Fixed.Eye * f.z);
  d.Fixed.Normal = normalize(a.Fixed.Normal * f.x + b.Fixed.Normal * f.y + c.Fixed.Normal * f.z);
  d.Fixed.MaterialNormal = normalize(a.Fixed.MaterialNormal * f.x + b.Fixed.MaterialNormal * f.y + c
      .Fixed.MaterialNormal * f.z);

  // Interpolate linear values.

  d.Color = a.Color * f.x + b.Color * f.y + c.Color * f.z;
  d.Fixed.MaterialCoords = a.Fixed.MaterialCoords * f.x + b.Fixed.MaterialCoords * f.y + c.Fixed.
      MaterialCoords * f.z;
  d.Fixed.Distance = a.Fixed.Distance * f.x + b.Fixed.Distance * f.y + c.Fixed.Distance * f.z;
  d.Fixed.Position = a.Fixed.Position * f.x + b.Fixed.Position * f.y + c.Fixed.Position * f.z;
  d.Fixed.Shadow = a.Fixed.Shadow * f.x + b.Fixed.Shadow * f.y + c.Fixed.Shadow * f.z;

  // Interpolate material token.

  if (all(a.Fixed.MaterialIds == b.Fixed.MaterialIds) && all(a.Fixed.MaterialIds == c.Fixed.
  MaterialIds))
  {
    // In most cases, the material IDs will be identical. In this case, just sum up the weights.

    d.Fixed.MaterialIds = a.Fixed.MaterialIds;
    d.WeightsA = a.WeightsA * f.x + b.WeightsA * f.y + c.WeightsA * f.z;
  }
  else
  {
    int4 idsA;
    int4 idsB;
    int4 idsC;
    float4 weightsA;
    float4 weightsB;
    float4 weightsC;
    float weightSum;
    int countA;
    int countB;
    int countC;

    // Need to mix three material tokens.

    idsA = a.Fixed.MaterialIds;
    idsB = b.Fixed.MaterialIds;
    idsC = c.Fixed.MaterialIds;
    weightsA = a.WeightsA;
    weightsB = b.WeightsA;
    weightsC = c.WeightsA;
    weightSum = dot(weightsA * f.x + weightsB * f.y + weightsC * f.z, 1.0);
    countA = (int) dot(sign(weightsA), 1.0);
    countB = (int) dot(sign(weightsB), 1.0);
    countC = (int) dot(sign(weightsC), 1.0);
    d.Fixed.MaterialIds = 0;
    d.WeightsA = 0.0;

    // Three material tokens can contain at most 12 distinct materials.

    while (max(countA, max(countB, countC)) > 0)
    {
      float weight;
      int id;

      weight = 0.0;
      id = max(idsA.x, max(idsB.x, idsC.x));

      if (idsA.x == id)
      {
        countA--;

        weight += weightsA.x * f.x;
        idsA = int4(idsA.yzw, 0);
        weightsA = float4(weightsA.yzw, 0);
      }

      if (idsB.x == id)
      {
        countB--;

        weight += weightsB.x * f.y;
        idsB = int4(idsB.yzw, 0);
        weightsB = float4(weightsB.yzw, 0);
      }

      if (idsC.x == id)
      {
        countC--;

        weight += weightsC.x * f.z;
        idsC = int4(idsC.yzw, 0);
        weightsC = float4(weightsC.yzw, 0);
      }

      if (weight > d.WeightsA.x)
      {
        d.Fixed.MaterialIds = int4(id, d.Fixed.MaterialIds.xyz);
        d.WeightsA = float4(weight, d.WeightsA.xyz);
      }
      else if (weight > d.WeightsA.y)
      {
        d.Fixed.MaterialIds.yzw = int3(id, d.Fixed.MaterialIds.yz);
        d.WeightsA.yzw = float3(weight, d.WeightsA.yz);
      }
      else if (weight > d.WeightsA.z)
      {
        d.Fixed.MaterialIds.zw = int2(id, d.Fixed.MaterialIds.z);
        d.WeightsA.zw = float2(weight, d.WeightsA.z);
      }
      else if (weight > d.WeightsA.w)
      {
        d.Fixed.MaterialIds.w = id;
        d.WeightsA.w = weight;
      }
    }

    // Scale weights in order to retain the weight sum. Otherwise, visual artefacts may appear,
    // for example slightly darkened areas because of weight sums less than one.

    d.WeightsA *= weightSum / dot(d.WeightsA, 1.0);
  }

  d.WeightsB = MaterialComputeWeights(d.Fixed.MaterialIds, d.WeightsA, g_materialIds);

  return d;
}

float2x2 MaterialRotation()
{
  float sin;
  float cos;

  sin = g_materialTransform.x;
  cos = g_materialTransform.y;

  return float2x2(cos, sin, -sin, cos);
}

Material MaterialSample(int n, float3 coords, float3 normal, float3 color)
{
  int id;
  Material result;
  float3 f;
  float2 planarX;
  float2 planarY;
  float2 planarZ;
  float4 tempX;
  float4 tempY;
  float4 tempZ;
  float4 temp;
  float3 tangentX;
  float3 tangentY;
  float3 tangentZ;
  float2 normalScale;

  // Compute contribution per texturing plane.

  id = (int) g_materialIds[n];
  f = g_materialSettings[id].y == 0.0 ? float3(0, 1, 0) : normal * normal;

  if (TERRAIN_MATERIAL_DEBUG == 1)
    f = float3(f.x, 0, 0);
  else if (TERRAIN_MATERIAL_DEBUG == 2)
    f = float3(0, f.y, 0);
  else if (TERRAIN_MATERIAL_DEBUG == 3)
    f = float3(0, 0, f.z);

  MaterialComputeCoords(id, coords, normal, planarX, planarY, planarZ);

  // Sample displacement map.

  if (TERRAIN_MATERIAL_DISPLACEMENT)
  {
    result.Height = f.x * g_materialSettings[id].w * (gxTextures2DArray(g_displacement, 1, id,
        planarX).r - 0.5) + f.y * g_materialSettings[id].z * (gxTextures2DArray(g_displacement, 0,
        id, planarY).r - 0.5) + f.z * g_materialSettings[id].w * (gxTextures2DArray(g_displacement,
        1, id, planarZ).r - 0.5);
  }
  else
    result.Height = 0.0;

  // Sample albedo/reflectivity map.

  tempX = gxTextures2D(g_material, Material_V_AR(n), planarX);
  tempY = gxTextures2D(g_material, Material_H_AR(n), planarY);
  tempZ = gxTextures2D(g_material, Material_V_AR(n), planarZ);
  tempX.rgb = MaterialAlbedoBlend(tempX.rgb, g_materialBlendV[n].z, g_materialBlendV[n].x, color,
      g_materialBlendV[n].w, g_materialBlendV[n].y);
  tempY.rgb = MaterialAlbedoBlend(tempY.rgb, g_materialBlendH[n].z, g_materialBlendH[n].x, color,
      g_materialBlendH[n].w, g_materialBlendH[n].y);
  tempZ.rgb = MaterialAlbedoBlend(tempZ.rgb, g_materialBlendV[n].z, g_materialBlendV[n].x, color,
      g_materialBlendV[n].w, g_materialBlendV[n].y);
  temp = tempX * f.x + tempY * f.y + tempZ * f.z;
  result.Albedo = temp.rgb;
  result.Reflectivity = 1.0 - temp.a;
  result.Normal = 0.0;

  // Setup tangent vectors. These will be used to construct the tangent-space for of the six planes.

  MaterialTangentSpace(normal, tangentX, tangentY, tangentZ);

  // Adjust normal vector in tangent space according to material displacement scale factor.
  // 
  // Scale tangent-space along up-vector:
  //      |   1   0   0   |
  // M  = |   0   1   0   |
  //      |   0   0   f   |
  // 
  // Normals are transformed with inverse transpose matrix:
  //      |   1   0   0   |
  // M' = |   0   1   0   |
  //      |   0   0 1/f   |

  normalScale = TERRAIN_MATERIAL_DISPLACEMENT ? max(1E-10, g_normalScale[n] * g_displacementSettings
      .x) : g_normalScale[n];

  // Texture plane X: sample normal/occlusion/roughness + specular map (vertical)

  temp = gxTextures2D(g_material, Material_V_NOR(n), planarX);
  result.Occlusion = f.x * temp.g;
  result.Gloss = f.x * temp.a;
  temp.xyz = DecodeNormalXY(temp.rb);
  temp.z /= normalScale.y;

  if (normal.x < 0.0)
  {
    // X+ (texture) => +Z (light-space)
    // Y+ (texture) => -Y (light-space)

    temp.xyz = mul(temp.xyz, float3x3(tangentZ, -tangentY, normal));
  }
  else
  {
    // X+ (texture) => -Z (light-space)
    // Y+ (texture) => -Y (light-space)

    temp.xyz = mul(temp.xyz, float3x3(-tangentZ, -tangentY, normal));
  }

  result.Normal += normalize(temp.xyz) * f.x;

  // Texture plane Z: sample normal/occlusion/roughness + specular map (vertical)

  temp = gxTextures2D(g_material, Material_V_NOR(n), planarZ);
  result.Occlusion += f.z * temp.g;
  result.Gloss += f.z * temp.a;
  temp.xyz = DecodeNormalXY(temp.rb);
  temp.z /= normalScale.y;

  if (normal.z < 0.0)
  {
    // X+ (texture) => -X (light-space)
    // Y+ (texture) => -Y (light-space)

    temp.xyz = mul(temp.xyz, float3x3(-tangentX, -tangentY, normal));
  }
  else
  {
    // X+ (texture) => +X (light-space)
    // Y+ (texture) => -Y (light-space)

    temp.xyz = mul(temp.xyz, float3x3(tangentX, -tangentY, normal));
  }

  result.Normal += normalize(temp.xyz) * f.z;

  // Texture plane Y: Sample normal/occlusion/roughness + specular map (horizontal)

  temp = gxTextures2D(g_material, Material_H_NOR(n), planarY);
  result.Occlusion += f.y * temp.g;
  result.Gloss += f.y * temp.a;
  temp.xyz = DecodeNormalXY(temp.rb);
  temp.z /= normalScale.x;

  if (normal.y < 0.0)
  {
    // X+ (texture) => -Z (light-space)
    // Y+ (texture) => -X (light-space)

    temp.xyz = mul(temp.xyz, float3x3(-tangentZ, -tangentX, normal));
  }
  else
  {
    // X+ (texture) => +X (light-space)
    // Y+ (texture) => +Z (light-space)

    temp.xyz = mul(temp.xyz, float3x3(tangentX, tangentZ, normal));
  }

  result.Normal += normalize(temp.xyz) * f.y;

  // Finalize values.

  result.Normal = normalize(result.Normal);
  result.Gloss = 1.0 - result.Gloss;

  // Apply debug flags.

  if (TERRAIN_MATERIAL_DEBUG != 0)
  {
    // | # | Color   |   Normal Sign   |
    // |   |         | R:X | G:Y | B:Z |
    // +---+---------+-----+-----+-----+
    // | 0 | Black   |  -  |  -  |  -  |
    // | 1 | Red     |  +  |  -  |  -  |
    // | 2 | Green   |  -  |  +  |  -  |
    // | 3 | Yellow  |  +  |  +  |  -  |
    // | 4 | Blue    |  -  |  -  |  +  |
    // | 5 | Magenta |  +  |  -  |  +  |
    // | 6 | Cyan    |  -  |  +  |  +  |
    // | 7 | White   |  +  |  +  |  +  |

    result.Gloss = 0.0;
    result.Reflectivity = 0.0;
    result.Albedo = FromSrgb(float3(sign(normal.x), sign(normal.y), sign(normal.z)) / 4.0 + 0.5);
  }

  return result;
}

void MaterialTangentSpace(float3 normal, out float3 tangentX, out float3 tangentY, out float3
    tangentZ)
{
  float xx;
  float xy;
  float xz;
  float yy;
  float yz;
  float zz;

  xx = normal.x * normal.x;
  xy = normal.x * normal.y;
  xz = normal.x * normal.z;
  yy = normal.y * normal.y;
  yz = normal.y * normal.z;
  zz = normal.z * normal.z;
  tangentX = normalize(float3(-zz - yy, xy, xz));
  tangentY = normalize(float3(xy, -xx - zz, yz));
  tangentZ = normalize(float3(xz, yz, -yy - xx));
}

float MaterialTessellationFactor(float size, float distance)
{
  float pixels;
  float factor;

  // Compute screen-size of triangle edge length.

  pixels = size * g_vertexBlend.x / distance;

  // Compute tessellation factor, based on subdivision scheme.

  factor = pixels / g_displacementSettings.y;

  // Scale with artificial displacement exaggeration factor.

  factor *= min(1.0, g_displacementSettings.x);

  // Fade out with layer distance range, to avoid Z-fighting artefacts with adjacent terrain layers.

  factor = DisplacementFade(factor, distance);

  return max(1.0, min(64.0, factor));
}

float MaterialTessellationFactor(GS_In_Material a0, GS_In_Material a1, DisplacementVarying b0, 
    DisplacementVarying b1)
{
  return MaterialTessellationFactor(sqrt(length(b0.Position - b1.Position) * (abs(b0.Displacement.y)
      + abs(b1.Displacement.y)) / 2.0), (a0.Fixed.Distance + a1.Fixed.Distance) / 2.0);
}

float TextureBlend(int atlas, float2 coords, float distance)
{
  float result;

  // Compute tile blend factor from texture atlas coefficients.

  result = dot(float4(1, coords.x, coords.y, coords.x * coords.y), g_tileBlend0[atlas] / distance +
      g_tileBlend1[atlas]);
  result = clamp(result, 0.0, 1.0);

  // Apply time-based tile blending.

  return result * g_tileBlendFactor[atlas];
}

float2 TextureClamp(float2 coordinates, float4 bounds)
{
  return clamp(coordinates, bounds.xz, bounds.yw);
}

void TextureCoords(int atlas, float3 texCoords, float3 texCoordsExtra, out float2 coordsBase, out 
    float2 coords)
{
  int idxBase;
  int idxTile;

  idxBase = Texture_N(0, atlas);
  idxTile = Texture_N(1, atlas);
  coordsBase = TextureCoords(texCoords, texCoordsExtra, g_tileOffset[idxBase], g_tileU[idxBase],
      g_tileV[idxBase]);
  coords = TextureCoords(texCoords, texCoordsExtra, g_tileOffset[idxTile], g_tileU[idxTile], g_tileV
      [idxTile]);
}

float2 TextureCoords(float3 texCoordsHigh, float3 texCoordsLow, float4 offset, float3 u, float3 v)
{
  return offset.xy + float2(dot(texCoordsHigh, u), dot(texCoordsHigh, v)) + offset.zw + float2(dot(
      texCoordsLow, u), dot(texCoordsLow, v));
}

Material TextureSample(float4 vertex, float4 coords[TextureCount], float distance, out float opacity
)
{
  Material result;
  float4 samples[TextureCount + 2];
  float4 v;

  // Set default material values.

  samples[TextureCount] = float4(0, 0, 0, 1);
  samples[TextureCount + 1] = vertex;
  result.Height = 0.0;

  // Fetch texture atlas samples.

  for (int i = 0; i < TextureCount; i++)
  {
    int idxBase;
    int idxTile;
    float4 sampleBase;
    float4 sampleTile;

    idxBase = Texture_N(0, i);
    idxTile = Texture_N(1, i);
    sampleBase = gxTextures2DArray(g_tileTexture, idxBase, g_tileSlice[idxBase], TextureClamp(coords
        [i].xy, g_tileBounds[idxBase]));
    sampleTile = gxTextures2DArray(g_tileTexture, idxTile, g_tileSlice[idxTile], TextureClamp(coords
        [i].zw, g_tileBounds[idxTile]));
    sampleBase.rgb = pow(sampleBase.rgb, g_tileGamma[idxBase]);
    sampleTile.rgb = pow(sampleTile.rgb, g_tileGamma[idxTile]);
    samples[i] = lerp(sampleBase, sampleTile, TextureBlend(i, coords[i].zw, distance));
  }

  // Albedo

  v = g_materialAlbedo;
  result.Albedo.r = samples[v.x][v.y];
  result.Albedo.g = samples[v.x][v.z];
  result.Albedo.b = samples[v.x][v.w];

  // Gloss

  result.Gloss = DecodeScalar(samples, g_materialGloss);

  // Normal

  v = g_materialNormal;
  result.Normal = float3(samples[v.x][v.y], samples[v.x][v.z], samples[v.x][v.w]);
  result.Normal = result.Normal * 2.0 - 1.0;
  result.Normal = Normalize(result.Normal, float3(0, 0, 0));

  // Occlusion

  result.Occlusion = DecodeScalar(samples, g_materialOcclusion);

  // Reflectivity

  result.Reflectivity = DecodeScalar(samples, g_materialReflectivity);

  // Opacity

  opacity = DecodeScalar(samples, g_materialOpacity);

  return result;
}

float VertexBlend(float2 levelOfDetail, float distanceSqr)
{
  float2 lodRange;

  // Compute screen-space area that corresponds to terrain-space area of vertex:
  // - Vertex mesh sector (base LOD)
  // - Vertex neighborhood (maximum LOD)

  lodRange = levelOfDetail * g_vertexBlend.y / distanceSqr;

  // Early-exit for the case when per-vertex level-of-detail is disabled.

  if (g_vertexBlend.w <= lodRange.y)
    return 1.0;

  // Compute normalized blend factor between base LOD and maximum LOD.

  lodRange = log2(lodRange);

  return saturate((lodRange.x - g_vertexBlend.w) / (lodRange.x - lodRange.y));
}

#endif
