// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Triangulate.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__TERRAIN__TRIANGULATE
#define TINMAN__TERRAIN__TRIANGULATE

#include "../../GpuX.hlsl"
#include "CB_Compute_Static.hlsl"
#include "SR_Triangulate.hlsl"

// -------------------------------------------------------------------------------------------------
// Constants
// -------------------------------------------------------------------------------------------------

static const int ComputeGroups = 16;

static const int ComputeThreads = 119;

static const int XDag_GMask = -1879048193;

static const int XDag_LL = 0;

static const int XDag_RL = 2;

// -------------------------------------------------------------------------------------------------
// Shader Variables
// -------------------------------------------------------------------------------------------------

groupshared int computeOutputEnd;

groupshared int computeOutputIdx;

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

void Compute_OutputDraw(int index, int first, int count);

void Compute_OutputTriangle(int a, int b, int c);

void Compute_SectorList(int from, int to, int step);

void Compute_TerminalTriangle(int from, int to, int step);

int XDag_A(int v);

int XDag_C(int v, int idx);

int XDag_G(int v);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

void Compute_OutputDraw(int index, int first, int count)
{
  index *= 5;

  // IndexCountPerInstance

  g_draw[index + 0] = count;

  // InstanceCount

  g_draw[index + 1] = 1;

  // StartIndexLocation

  g_draw[index + 2] = first;

  // BaseVertexLocation

  g_draw[index + 3] = 0;

  // StartInstanceLocation

  g_draw[index + 4] = 0;
}

void Compute_OutputTriangle(int a, int b, int c)
{
  int index;

  InterlockedAdd(computeOutputIdx, 3, index);

  if (index < computeOutputEnd)
  {
    g_output[index + 0] = a;
    g_output[index + 1] = b;
    g_output[index + 2] = c;
  }
}

void Compute_SectorList(int from, int to, int step)
{
  for (int index = from; index < to; index += step)
  {
    int code;
    int ci;
    bool winding;
    int v;
    int g;
    int gi;
    int4 lgra;
    int temp;

    // Decode sector code.

    code = g_input[index];
    ci = code & 7;
    winding = (code & 8) != 0 ^ g_flip;
    v = code >> 4 & 0xFFFFFFF;
    g = XDag_G(v);
    gi = g >> 29 & 3;
    g &= XDag_GMask;
    lgra = int4(XDag_C(g, -gi & 3), g, XDag_C(g, gi ^ 1), XDag_A(v));

    // Leaf sector quadrant?

    if (ci < 4)
    {
      int c0;
      int c1;
      int p;

      c0 = XDag_C(v, -ci & 3);

      if (c0 < 0)
        c0 = lgra[ci + 1 & 3];

      c1 = XDag_C(v, ci ^ 1);
      p = lgra[ci];

      if (winding)
      {
        temp = p;
        p = v;
        v = temp;
      }

      Compute_OutputTriangle(v, p, c0);

      if (c1 >= 0)
        Compute_OutputTriangle(p, v, c1);
    }
    else
    {
      int stack[129];
      int pos;

      // Initialize state machine for mesh traversal.

      pos = -1;
      stack[++pos] = lgra.z;
      stack[++pos] = v;
      stack[++pos] = XDag_RL;
      stack[++pos] = lgra.x;
      stack[++pos] = v;
      stack[++pos] = XDag_LL;

      // Triangulate non-leaf quadrant.

      while (pos >= 2)
      {
        int n;
        int c0;
        int c1;
        int p0;

        n = stack[pos--];
        v = stack[pos--];
        p0 = stack[pos--];
        c0 = XDag_C(v, n);
        c1 = XDag_C(v, n ^ 1);

        // Non-terminal triangle?

        if (c0 >= 0 && c1 >= 0)
        {
          stack[++pos] = v;
          stack[++pos] = c0;
          stack[++pos] = XDag_LL;
          stack[++pos] = v;
          stack[++pos] = c1;
          stack[++pos] = XDag_RL;
        }
        else
        {
          bool n0;
          int p1;
          int p2;

          // Leaf-sector.

          n0 = n == 0;

          if (n0 ^ winding)
          {
            p1 = XDag_G(v) & XDag_GMask;
            p2 = XDag_A(v);
          }
          else
          {
            p1 = XDag_A(v);
            p2 = XDag_G(v) & XDag_GMask;
          }

          // Minimal leaf-sector with two triangles?

          if (c0 < 0 && c1 < 0)
          {
            Compute_OutputTriangle(v, p0, p1);
            Compute_OutputTriangle(v, p2, p0);
          }
          else
          {
            // Leaf-sector with more than two triangles.

            n = 2;

            if (winding)
            {
              temp = c0;
              c0 = c1;
              c1 = temp;
              n ^= 3;
            }

            if (c0 >= 0)
            {
              Compute_OutputTriangle(v, p0, c0);
              Compute_OutputTriangle(v, p2, p0);
            }
            else
            {
              Compute_OutputTriangle(v, c1, p0);
              Compute_OutputTriangle(v, p0, p1);

              c0 = c1;
              p1 = p2;
              temp = XDag_C(c0, n);

              if (temp < 0)
              {
                Compute_OutputTriangle(v, p1, c0);

                continue;
              }

              Compute_OutputTriangle(v, temp, c0);

              v = c0;
              c0 = temp;
            }

            c1 = p1;
            temp = XDag_C(c0, n ^ 3);

            while (temp >= 0)
            {
              Compute_OutputTriangle(v, c0, temp);

              v = c0;
              c0 = temp;
              temp = XDag_C(c0, n);

              if (temp >= 0)
              {
                Compute_OutputTriangle(v, temp, c0);

                v = c0;
                c0 = temp;
                temp = XDag_C(c0, n ^ 3);
              }
              else
              {
                c1 = c0;
                c0 = p1;
              }
            }

            Compute_OutputTriangle(v, c0, c1);
          }
        }
      }
    }
  }
}

void Compute_TerminalTriangle(int from, int to, int step)
{
  for (int index = from; index < to; index += step)
  {
    int code;
    bool winding;
    int g;
    int p;
    int v;
    int ha;
    int hb;
    int c0;
    int c1;
    int idx;

    // Decode terminal triangle code.

    code = g_input[index];
    g = code >> 6 & 0x3FFFFFF;
    p = XDag_C(g, code >> 1 & 3);
    v = XDag_C(p, code & 3);
    winding = (code & 32) != 0 ^ g_flip;

    // Merge states to reduce triangulation cases.

    if ((code & 1) == 0)
    {
      // P.?L == V

      ha = (code & 8) != 0 ? 0 : 1;
      hb = (code & 16) != 0 ? 2 : 1;
      c0 = g;
      c1 = XDag_A(v);
      idx = XDag_LL;
    }
    else
    {
      // P.?R == V

      ha = (code & 16) != 0 ? 0 : 1;
      hb = (code & 8) != 0 ? 2 : 1;
      c0 = XDag_A(v);
      c1 = g;
      idx = XDag_RL;
    }

    // Decode halves of terminal triangle separately.

    for (int h = ha; h < hb; h++)
    {
      int ch;
      int n;
      int c;
      int v1;
      int v2;
      int v3;
      bool h1;

      if (h == 0)
        ch = c0;
      else
        ch = c1;

      n = XDag_RL ^ h * 3;
      c = XDag_C(v, idx ^ h);
      v2 = v;
      v3 = p;
      h1 = h != 0;

      if (h1 ^ winding)
      {
        if (c < 0)
        {
          Compute_OutputTriangle(v3, v2, ch);

          continue;
        }

        v1 = c;

        Compute_OutputTriangle(v3, v2, v1);

        c = XDag_C(c, n ^= 3);
        v3 = v2;
        v2 = v1;
      }

      while (c >= 0)
      {
        v1 = c;

        Compute_OutputTriangle(v1, v2, v3);

        c = XDag_C(c, n ^ 3);
        v3 = v2;
        v2 = v1;

        if (c >= 0)
        {
          v1 = c;

          Compute_OutputTriangle(v3, v2, v1);

          c = XDag_C(c, n);
          v3 = v2;
          v2 = v1;
        }
        else
        {
          v1 = v3;
          v3 = ch;
          ch = v1;
        }
      }

      Compute_OutputTriangle(ch, v2, v3);
    }
  }
}

int XDag_A(int v)
{
  return g_structure[v]._2_A;
}

int XDag_C(int v, int idx)
{
  return g_structure[v]._0_C[idx];
}

int XDag_G(int v)
{
  return g_structure[v]._1_G;
}

#endif
